import pytest
from fastapi import Depends, FastAPI
from fastapi.testclient import TestClient

from farl import AsyncFarl, Farl
from farl.dependencies import rate_limit, rate_limits
from farl.exceptions import QuotaExceeded


app = FastAPI()


@pytest.fixture
def api():
    with TestClient(app) as client:
        yield client


@app.get(
    "/unset-farl",
    dependencies=[Depends(rate_limit({"amount": 1}))],
)
def unset_farl(): ...


def test_unset_farl(api: TestClient):
    with pytest.raises(ValueError, match="farl instance is required") as excinfo:
        api.get("/unset-farl")
    assert excinfo.type is ValueError


@app.get(
    "/raise-exc",
    dependencies=[
        Depends(
            rate_limit(
                {"amount": 1},
                farl=Farl(),
            )
        )
    ],
)
def raise_exc(): ...


def test_raise_exc(api: TestClient):
    res = api.get("/raise-exc")
    assert res.is_success

    with pytest.raises(QuotaExceeded) as excinfo:
        api.get("/raise-exc")

    assert excinfo.type is QuotaExceeded
    assert excinfo.value.data.model_dump()["violated-policies"] == ["preminute"]


@app.get(
    "/multiple-ratelimit",
    dependencies=[
        Depends(
            rate_limit(
                [
                    {"amount": 1},
                    {"amount": 2, "multiples": 3, "time": "second"},
                ],
                farl=Farl(),
                policy_name="multiple",
            )
        )
    ],
)
def multiple_ratelimit(): ...


def test_multiple_ratelimit(api: TestClient):
    api.get("/multiple-ratelimit")
    with pytest.raises(QuotaExceeded) as excinfo:
        api.get("/multiple-ratelimit")

    assert excinfo.type is QuotaExceeded
    assert excinfo.value.data.model_dump()["violated-policies"] == [
        "multiple-preminute"
    ]


@app.get(
    "/rate-limits",
    dependencies=[
        Depends(
            rate_limits(
                rate_limit(
                    {"amount": 1},
                    farl=Farl(namespace="test"),
                    policy_name="a",
                    error_class=None,
                ),
                rate_limit(
                    [
                        {"amount": 1, "time": "month"},
                        {"amount": 2, "multiples": 3, "time": "second"},
                    ],
                    farl=AsyncFarl(),
                    policy_name="b",
                    error_class=None,
                ),
            )
        )
    ],
)
def _rate_limits(): ...


def test_rate_limits(api: TestClient):
    api.get("/rate-limits")
    with pytest.raises(QuotaExceeded) as excinfo:
        api.get("/rate-limits")

    assert excinfo.type is QuotaExceeded
    assert excinfo.value.data.model_dump()["violated-policies"] == ["a", "b-premonth"]


@app.get(
    "/mismatched-costs-partition-keys",
    dependencies=[
        Depends(
            rate_limit(
                {"amount": 1},
                farl=Farl(),
                get_partition_key=lambda: ["key1", "key2"],
                get_cost=lambda: [1],  # Only one cost for two partition keys
            )
        )
    ],
)
def mismatched_costs_partition_keys(): ...


def test_mismatched_costs_partition_keys(api: TestClient):
    with pytest.raises(
        ValueError,
        match=r"Number of costs \(1\) must match number of partition keys \(2\)",
    ):
        api.get("/mismatched-costs-partition-keys")


@app.get(
    "/mismatched-partition-keys-costs",
    dependencies=[
        Depends(
            rate_limit(
                {"amount": 1},
                farl=Farl(),
                get_partition_key=lambda: ["key1"],
                get_cost=lambda: [1, 2, 3],  # Three costs for one partition key
            )
        )
    ],
)
def mismatched_partition_keys_costs(): ...


def test_mismatched_partition_keys_costs(api: TestClient):
    with pytest.raises(
        ValueError,
        match=r"Number of costs \(3\) must match number of partition keys \(1\)",
    ):
        api.get("/mismatched-partition-keys-costs")



from fastapi import FastAPI
from fastapi.testclient import TestClient

from farl.exceptions import (
    AbnormalUsageDetected,
    FarlError,
    QuotaExceeded,
    TemporaryReducedCapacity,
    farl_exceptions_handler,
)


app = FastAPI()


quota_exceeded_exc = QuotaExceeded(violated_policies=["preMinute"])


@app.get("/quota-exceeded")
def quota_exceeded():
    raise quota_exceeded_exc


abnormal_usage_detected_exc = AbnormalUsageDetected(
    violated_policies=["preMinute", "preHour"]
)


@app.get("/abnormal-usage-detected")
def abnormal_usage_detected():
    raise abnormal_usage_detected_exc


temporary_reduced_capacity_exc = TemporaryReducedCapacity(
    violated_policies=["preMinute"]
)


@app.get("/temporary-reduced-capacity")
def temporary_reduced_capacity():
    raise temporary_reduced_capacity_exc


app.add_exception_handler(FarlError, farl_exceptions_handler)


def test_exceptions():
    with TestClient(app) as client:
        res = client.get("/quota-exceeded")
        assert res.is_error
        assert res.headers["content-type"] == quota_exceeded_exc.media_type
        res_data: dict = res.json()
        assert res_data["type"] == quota_exceeded_exc.type
        assert res_data["title"] == quota_exceeded_exc.title
        assert res.status_code == quota_exceeded_exc.status
        assert (
            res_data["violated-policies"] == quota_exceeded_exc.data.violated_policies  # pyright: ignore [reportAttributeAccessIssue]
        )

        res = client.get("/abnormal-usage-detected")
        assert res.is_error
        assert res.headers["content-type"] == abnormal_usage_detected_exc.media_type
        res_data: dict = res.json()
        assert res_data["type"] == abnormal_usage_detected_exc.type
        assert res_data["title"] == abnormal_usage_detected_exc.title
        assert res.status_code == abnormal_usage_detected_exc.status
        assert (
            res_data["violated-policies"]
            == abnormal_usage_detected_exc.data.violated_policies  # pyright: ignore [reportAttributeAccessIssue]
        )

        res = client.get("/temporary-reduced-capacity")
        assert res.is_error
        assert res.headers["content-type"] == temporary_reduced_capacity_exc.media_type
        res_data: dict = res.json()
        assert res_data["type"] == temporary_reduced_capacity_exc.type
        assert res_data["title"] == temporary_reduced_capacity_exc.title
        assert res.status_code == temporary_reduced_capacity_exc.status
        assert (
            res_data["violated-policies"]
            == temporary_reduced_capacity_exc.data.violated_policies  # pyright: ignore [reportAttributeAccessIssue]
        )



from fastapi import FastAPI, Request
from fastapi.testclient import TestClient

from farl.constants import HEADER_RATELIMIT, HEADER_RATELIMIT_POLICY, STATE_KEY
from farl.middleware import FarlMiddleware
from farl.utils import FarlState, HeaderRateLimit, HeaderRateLimitPolicy


app = FastAPI()
app.add_middleware(FarlMiddleware)


a_policy = HeaderRateLimitPolicy("testvalue", 10, None, 60, "/a:GET")
a_r = HeaderRateLimit(a_policy.policy, 1, 23, a_policy.partition_key)


@app.get("/a")
def a(request: Request):
    state: dict = request.scope["state"]
    farl_state: FarlState = state[STATE_KEY]

    farl_state["policy"].append(a_policy)
    farl_state["state"].append(a_r)


b_policy = [
    HeaderRateLimitPolicy("testvalue-preMinute", 10, None, 60, "/b:GET"),
    HeaderRateLimitPolicy("testvalue-preHour", 10, None, 60 * 60, "/b:GET"),
]
b_r = [
    HeaderRateLimit(b_policy[0].policy, 6, 5, b_policy[0].partition_key),
    HeaderRateLimit(b_policy[1].policy, 3, 5, b_policy[1].partition_key),
]


@app.get("/b")
async def b(request: Request):
    state: dict = request.scope["state"]
    farl_state: FarlState = state[STATE_KEY]
    farl_state["policy"].extend(b_policy)
    farl_state["state"].extend(b_r)


c_policy = [
    HeaderRateLimitPolicy("testvalue-preMinute", 10, None, 60, "/c:GET"),
    HeaderRateLimitPolicy("testvalue-preHour", 10, None, 60 * 60, "/c:GET"),
    HeaderRateLimitPolicy("testvalue-preDay", 10, None, 60 * 60 * 24, "/c:GET"),
]
c_r = [
    HeaderRateLimit(c_policy[0].policy, 6, 2, c_policy[0].partition_key),
    HeaderRateLimit(c_policy[1].policy, 3, 10, c_policy[1].partition_key),
    HeaderRateLimit(c_policy[2].policy, 1, None, c_policy[2].partition_key),
]


@app.get("/c")
async def c(request: Request):
    state: dict = request.scope["state"]
    farl_state: FarlState = state[STATE_KEY]
    farl_state["policy"].extend(c_policy)
    farl_state["state"].extend(c_r)


d_policy = [
    HeaderRateLimitPolicy("testvalue-preMinute", 10, None, 60, "/d:GET"),
    HeaderRateLimitPolicy("testvalue-preHour", 10, None, 60 * 60, "/d:GET"),
    HeaderRateLimitPolicy("testvalue-preDay", 10, None, 60 * 60 * 24, "/d:GET"),
]
d_r = [
    HeaderRateLimit(d_policy[0].policy, 6, None, d_policy[0].partition_key),
    HeaderRateLimit(d_policy[1].policy, 3, None, d_policy[1].partition_key),
    HeaderRateLimit(d_policy[2].policy, 1, None, d_policy[2].partition_key),
]


@app.get("/d")
async def d(request: Request):
    state: dict = request.scope["state"]
    farl_state: FarlState = state[STATE_KEY]
    farl_state["policy"].extend(d_policy)
    farl_state["state"].extend(d_r)


def test_middleware():
    with TestClient(app) as client:
        a_res = client.get("/a")
        assert a_res.is_success
        assert a_res.headers[HEADER_RATELIMIT_POLICY] == str(a_policy)
        assert a_res.headers[HEADER_RATELIMIT] == str(a_r)

        b_res = client.get("/b")
        assert b_res.is_success
        b_value = ",".join(str(i) for i in b_policy)
        assert b_res.headers[HEADER_RATELIMIT_POLICY] == b_value
        assert b_res.headers[HEADER_RATELIMIT] == str(b_r[1])

        c_res = client.get("/c")
        assert c_res.is_success
        c_value = ",".join(str(i) for i in c_policy)
        assert c_res.headers[HEADER_RATELIMIT_POLICY] == c_value
        assert c_res.headers[HEADER_RATELIMIT] == str(c_r[1])

        d_res = client.get("/d")
        assert d_res.is_success
        d_value = ",".join(str(i) for i in d_policy)
        assert d_res.headers[HEADER_RATELIMIT_POLICY] == d_value
        assert d_res.headers[HEADER_RATELIMIT] == str(d_r[2])



from typing import cast

from starlette.datastructures import MutableHeaders

from farl.constants import HEADER_RATELIMIT, HEADER_RATELIMIT_POLICY, STATE_KEY
from farl.utils import FarlState

from .types import AsyncFarlProtocol, FarlProtocol


class FarlMiddleware:
    def __init__(
        self,
        app,
        farl: FarlProtocol | AsyncFarlProtocol | None = None,
    ) -> None:
        self.app = app
        self.farl = farl

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        farl_state = (
            FarlState(policy=[], state=[], violated=[])
            if self.farl is None
            else FarlState(farl=self.farl, policy=[], state=[], violated=[])
        )
        scope.setdefault("state", {}).setdefault(STATE_KEY, farl_state)

        async def send_wrapper(message) -> None:
            if message["type"] == "http.response.start" and (
                farl_state := scope.get("state", {}).get(STATE_KEY)
            ):
                farl_state = cast(FarlState, farl_state)

                best_none = None
                best = None
                for i in farl_state["state"]:
                    if i.reset_timestamp is None:
                        if (best_none is None) or (i.remaining < best_none.remaining):
                            best_none = i
                    elif (best is None) or (
                        i.remaining,
                        i.reset_timestamp,
                    ) < (
                        best.remaining,
                        best.reset_timestamp,
                    ):
                        best = i

                headers = MutableHeaders(scope=message)

                if farl_state["policy"]:
                    ratelimit_policy = ",".join(str(i) for i in farl_state["policy"])
                    headers.setdefault(HEADER_RATELIMIT_POLICY, ratelimit_policy)

                if best or best_none:
                    ratelimit = str(best or best_none)
                    headers.setdefault(HEADER_RATELIMIT, ratelimit)

                await send(message)
            else:
                await send(message)

        await self.app(scope, receive, send_wrapper)



from collections.abc import Awaitable, Callable, Sequence
from typing import Protocol, TypeVar

import limits
import limits.aio
from fastapi import Request
from limits import RateLimitItem
from pydantic import networks

from farl.utils import RateLimitDictValue, RateLimitTimeArg


Key = str
KeyResult = Key | Sequence[Key]
GetKeyDependency = Callable[..., KeyResult | Awaitable[KeyResult]]
GetRequestKey = Callable[[Request], KeyResult | Awaitable[KeyResult]]


Cost = int
CostResult = Cost | Sequence[Cost]
GetCostDependency = Callable[..., CostResult | Awaitable[CostResult]]
GetRequestCost = Callable[[Request], CostResult | Awaitable[CostResult]]


RateLimitPolicy = (
    str
    | RateLimitItem
    | Sequence[RateLimitItem]
    | RateLimitTimeArg
    | Sequence[RateLimitTimeArg]
    | RateLimitDictValue
    | Sequence[RateLimitDictValue]
)
GetRateLimitPolicyDependency = Callable[
    ...,
    RateLimitPolicy | Awaitable[RateLimitPolicy],
]
GetRequestRateLimitPolicy = Callable[
    [Request],
    RateLimitPolicy | Awaitable[RateLimitPolicy],
]


class RedisDsn(networks.RedisDsn):
    _constraints = networks.UrlConstraints(
        allowed_schemes=[
            "redis",
            "rediss",
            "redis+sentinel",
            "redis+cluster",
        ],
        default_host="localhost",
        default_port=6379,
        default_path="/0",
        host_required=True,
    )


_T = TypeVar("_T")


class _FarlProtocol(Protocol[_T]):
    limiter: _T
    namespace: str | None
    key: KeyResult | GetRequestKey | None
    cost: CostResult | GetRequestCost | None
    policy: RateLimitPolicy | GetRequestRateLimitPolicy | None


FarlProtocol = _FarlProtocol[limits.strategies.RateLimiter]
AsyncFarlProtocol = _FarlProtocol[limits.aio.strategies.RateLimiter]
AnyFarlProtocol = FarlProtocol | AsyncFarlProtocol



from fastapi import Request


def request_endpoint(request: Request):
    request_path = request.scope.get("path") or ""
    request_method = request.method
    return f"{request_path}:{request_method}"



from .base import AsyncFarl, Farl
from .dependencies import rate_limit, rate_limits
from .exceptions import FarlError, farl_exceptions_handler
from .middleware import FarlMiddleware


__all__ = [
    "AsyncFarl",
    "Farl",
    "FarlError",
    "FarlMiddleware",
    "farl_exceptions_handler",
    "rate_limit",
    "rate_limits",
]



from typing import Literal

from fastapi import Request
from fastapi.responses import JSONResponse
from pydantic import Field, create_model


class FarlError(Exception):
    status: int
    type: str
    title: str
    media_type: str = "application/problem+json"

    @classmethod
    def _build_schema(cls):
        cls.model = create_model(
            cls.__name__,
            type=(Literal[cls.type], ...),
            title=(Literal[cls.title], ...),
            violated_policies=(list[str], Field(..., alias="violated-policies")),
            __config__={"serialize_by_alias": True, "validate_by_alias": False},
        )

    def __init_subclass__(cls) -> None:
        cls._build_schema()

    def __init__(
        self,
        *,
        violated_policies: list[str] = [],  # noqa: B006
        headers: dict | None = None,
    ) -> None:
        self.data = self.model(
            type=self.type,
            title=self.title,
            violated_policies=violated_policies,
        )
        self.headers = headers or {}


# 超出配额
# 如果服务器希望通知客户端其发送的请求超出了一个或多个配额策略，则可以使用此问题类型。
class QuotaExceeded(FarlError):
    status = 429

    type = "https://iana.org/assignments/http-problem-types#quota-exceeded"
    title = "Request cannot be satisifed as assigned quota has been exceeded"


# 暂时减少容量
# 如果服务器想要告知客户端，由于服务限制导致容量暂时减少，
# 客户端发送的请求超出当前无法满足的范围，则可以使用此问题类型。
# 服务器可以选择包含一个 RateLimit-Policy 字段，指示新的暂时较低的配额。
# 此问题类型将扩展成员“violated-policies”定义为一个字符串数组，其值为超出配额的策略名称
class TemporaryReducedCapacity(FarlError):
    status = 503

    type = "https://iana.org/assignments/http-problem-types#temporary-reduced-capacity"
    title = "Request cannot be satisifed due to temporary server capacity constraints"


class AbnormalUsageDetected(FarlError):
    status = 429

    type = "https://iana.org/assignments/http-problem-types#abnormal-usage-detected"
    title = "Request not satisifed due to detection of abnormal request pattern"


def farl_exceptions_handler(_request: Request, exc):
    return JSONResponse(
        exc.data.model_dump(exclude_unset=True),
        status_code=exc.status,
        media_type=exc.media_type,
        headers=exc.headers,
    )



import asyncio
import inspect
import logging
from collections.abc import Callable
from copy import copy

from fastapi import Depends, Request
from limits import RateLimitItem

from farl.constants import STATE_KEY
from farl.exceptions import FarlError, QuotaExceeded
from farl.types import (
    AnyFarlProtocol,
    Cost,
    CostResult,
    GetCostDependency,
    GetKeyDependency,
    GetRateLimitPolicyDependency,
    KeyResult,
    RateLimitPolicy,
)
from farl.utils import (
    FarlState,
    HeaderRateLimit,
    HeaderRateLimitPolicy,
    parse_rate_limit_value,
)

from .utils import request_endpoint


logger = logging.getLogger("farl")


def _get_policy_name(value: RateLimitItem):
    if value.multiples != 1:
        return f"pre{value.multiples}{value.GRANULARITY.name}"
    return f"pre{value.GRANULARITY.name}"


def _update_namespace(
    namespace: str | None,
    value: RateLimitItem,
):
    if namespace:
        result = copy(value)
        result.namespace = namespace
    else:
        result = value

    return value


async def _handle(
    *,
    farl: AnyFarlProtocol,
    farl_state: FarlState,
    policy_name: str,
    value: RateLimitItem,
    quota_unit: str | None,
    partition_key: str,
    keys: list,
    cost: int,
):
    limiter = farl.limiter
    farl_state["policy"].append(
        HeaderRateLimitPolicy(
            policy_name,
            value.amount,
            quota_unit,
            value.multiples * value.GRANULARITY.seconds,
            partition_key,
        )
    )

    hit_result = limiter.hit(value, *keys, cost=cost)
    if inspect.isawaitable(hit_result):
        hit_result = await hit_result

    stats_result = limiter.get_window_stats(value, *keys)
    if inspect.isawaitable(stats_result):
        stats_result = await stats_result

    ratelimit = HeaderRateLimit(
        policy_name,
        stats_result.remaining,
        stats_result.reset_time,
        partition_key,
    )

    farl_state["state"].append(ratelimit)

    if hit_result is False:
        logger.warning(
            "Rate limit exceeded for partition_key: %s, keys: %s, cost: %s",
            partition_key,
            keys,
            cost,
        )
        farl_state["violated"].append(ratelimit)


def _value(v):
    def _():
        return v

    return _


async def _get_keys(
    request: Request,
    dep_key: KeyResult | None,
    farl: AnyFarlProtocol,
    partition_key: str,
):
    keys: list[str] = []
    if farl.key is not None:
        k = farl.key
        if callable(k):
            k = k(request)
            if inspect.isawaitable(k):
                k = await k

        keys.extend([k] if isinstance(k, str) else k)

    keys.extend(
        ([dep_key] if isinstance(dep_key, str) else list(dep_key)) if dep_key else []
    )
    keys.append(partition_key)
    return keys


async def _get_cost(
    request: Request,
    dep_cost: CostResult | None,
    farl: AnyFarlProtocol,
):
    result = dep_cost
    if result is None:
        if farl.cost:
            cost_result = farl.cost
            if callable(cost_result):
                cost_result = cost_result(request)
                if inspect.isawaitable(cost_result):
                    cost_result = await cost_result
            result = cost_result
        else:
            result = 1

    return result


async def _check_partition(
    *,
    request: Request,
    key: KeyResult | None,
    farl: AnyFarlProtocol,
    partition_key: str,
    cost: Cost,
    value: RateLimitPolicy,
    farl_state: FarlState,
    default_policy_name: str | None,
    quota_unit: str | None,
):
    keys = await _get_keys(request, key, farl, partition_key)

    limit_values = parse_rate_limit_value(value)

    if len(limit_values) == 1:
        value = _update_namespace(farl.namespace, limit_values[0])

        policy_name = default_policy_name or _get_policy_name(value)

        await _handle(
            farl=farl,
            farl_state=farl_state,
            policy_name=policy_name,
            value=value,
            quota_unit=quota_unit,
            partition_key=partition_key,
            keys=keys,
            cost=cost,
        )
    else:
        for i in limit_values:
            value = _update_namespace(farl.namespace, i)

            policy_name = _get_policy_name(value)
            if default_policy_name is not None:
                policy_name = f"{default_policy_name}-{policy_name}"

            await _handle(
                farl=farl,
                farl_state=farl_state,
                policy_name=policy_name,
                value=value,
                quota_unit=quota_unit,
                partition_key=partition_key,
                keys=keys,
                cost=cost,
            )


def rate_limit(
    policy: RateLimitPolicy | GetRateLimitPolicyDependency,
    *,
    policy_name: str | None = None,
    quota_unit: str | None = None,
    get_key: KeyResult | GetKeyDependency | None = None,
    get_partition_key: KeyResult | GetKeyDependency = request_endpoint,
    get_cost: CostResult | GetCostDependency | None = None,
    error_class: type[FarlError] | None = QuotaExceeded,
    farl: AnyFarlProtocol | None = None,
):
    default_policy_name = policy_name

    value_dep = policy if callable(policy) else _value(policy)
    key_dep = get_key if callable(get_key) else _value(get_key)
    partition_key_dep = (
        get_partition_key if callable(get_partition_key) else _value(get_partition_key)
    )
    cost_dep = get_cost if callable(get_cost) else _value(get_cost)

    async def dependency(
        request: Request,
        value: RateLimitPolicy = Depends(value_dep),
        partition_key: KeyResult = Depends(partition_key_dep),
        key: KeyResult | None = Depends(key_dep),
        cost: CostResult | None = Depends(cost_dep),
    ):
        state = request.scope.setdefault("state", {})
        farl_state: FarlState = state.setdefault(
            STATE_KEY,
            FarlState(policy=[], state=[], violated=[]),
        )
        _farl: AnyFarlProtocol | None = farl or farl_state.get("farl")
        if _farl is None:
            raise ValueError("farl instance is required")

        if isinstance(partition_key, str):
            partition_key = [partition_key]

        costs = await _get_cost(request, cost, _farl)

        if isinstance(costs, int):
            costs = [costs for _ in partition_key]
        elif cost is None:
            costs = [1 for _ in partition_key]

        partition_keys = partition_key

        # Validate that costs and partition_keys have the same length
        if len(costs) != len(partition_keys):
            raise ValueError(
                f"Number of costs ({len(costs)}) must match number "
                f"of partition keys ({len(partition_keys)})"
            )

        await asyncio.gather(
            *[
                _check_partition(
                    request=request,
                    key=key,
                    farl=_farl,
                    partition_key=partition_key,
                    cost=cost,
                    value=value,
                    farl_state=farl_state,
                    default_policy_name=default_policy_name,
                    quota_unit=quota_unit,
                )
                for partition_key, cost in zip(partition_keys, costs, strict=True)
            ]
        )

        if farl_state["violated"] and error_class is not None:
            raise error_class(
                violated_policies=[i.policy for i in farl_state["violated"]]
            )

    return dependency


def rate_limits(*args: Callable, error_class: type[FarlError] = QuotaExceeded):
    def dependency(request: Request, **_ratelimit_deps):
        state: dict = request.scope.setdefault("state", {})
        farl_state: FarlState = state.setdefault(
            STATE_KEY,
            FarlState(policy=[], state=[], violated=[]),
        )
        if farl_state["violated"]:
            raise error_class(
                violated_policies=[i.policy for i in farl_state["violated"]]
            )

    sign = inspect.signature(dependency)
    param_mapping = sign.parameters.copy()
    param_mapping.pop("_ratelimit_deps")
    params = list(param_mapping.values())
    params.extend(
        inspect.Parameter(
            name=f"_ratelimit_dep_{index}",
            kind=inspect.Parameter.KEYWORD_ONLY,
            default=Depends(i),
        )
        for index, i in enumerate(args)
    )

    new_sign = sign.replace(parameters=params)
    setattr(dependency, "__signature__", new_sign)  # noqa: B010
    return dependency



from typing import Literal, cast, overload

import limits
import limits.aio
from pydantic import RedisDsn

from farl.types import (
    AsyncFarlProtocol,
    CostResult,
    FarlProtocol,
    GetRequestCost,
    GetRequestKey,
    GetRequestRateLimitPolicy,
    KeyResult,
    RateLimitPolicy,
)


try:
    from redis import ConnectionPool as RedisConnectionPool
    from redis.asyncio import ConnectionPool as AsyncRedisConnectionPool
except ImportError:
    RedisConnectionPool = None
    AsyncRedisConnectionPool = None


class Farl(FarlProtocol):
    @overload
    def __init__(
        self,
        *,
        redis_uri: str | RedisDsn | None,
        redis_connection_pool: "RedisConnectionPool | None" = None,  # pyright: ignore [reportInvalidTypeForm]
        redis_key_prefix: str = limits.storage.RedisStorage.PREFIX,
        redis_wrap_exceptions: bool = False,
        redis_options: dict | None = None,
        strategy: Literal[
            "fixed-window",
            "moving-window",
            "sliding-window-counter",
        ] = "fixed-window",
        namespace: str | None = None,
        rate_limit_key: KeyResult | GetRequestKey | None = None,
        rate_limit_cost: CostResult | GetRequestCost | None = None,
        rate_limit_policy: RateLimitPolicy | GetRequestRateLimitPolicy | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        *,
        storage_uri: str | RedisDsn,
        storage_options: dict | None = None,
        strategy: Literal[
            "fixed-window",
            "moving-window",
            "sliding-window-counter",
        ] = "fixed-window",
        namespace: str | None = None,
        rate_limit_key: KeyResult | GetRequestKey | None = None,
        rate_limit_cost: CostResult | GetRequestCost | None = None,
        rate_limit_policy: RateLimitPolicy | GetRequestRateLimitPolicy | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        *,
        strategy: Literal[
            "fixed-window",
            "moving-window",
            "sliding-window-counter",
        ] = "fixed-window",
        namespace: str | None = None,
        rate_limit_key: KeyResult | GetRequestKey | None = None,
        rate_limit_cost: CostResult | GetRequestCost | None = None,
        rate_limit_policy: RateLimitPolicy | GetRequestRateLimitPolicy | None = None,
    ) -> None: ...

    def __init__(
        self,
        *,
        redis_uri: str | RedisDsn | None = None,
        redis_connection_pool: "RedisConnectionPool | None" = None,  # pyright: ignore [reportInvalidTypeForm]
        redis_key_prefix: str = limits.storage.RedisStorage.PREFIX,
        redis_wrap_exceptions: bool = False,
        redis_options: dict | None = None,
        storage_uri: str | RedisDsn | None = None,
        storage_options: dict | None = None,
        strategy: Literal[
            "fixed-window",
            "moving-window",
            "sliding-window-counter",
        ] = "fixed-window",
        namespace: str | None = None,
        rate_limit_key: KeyResult | GetRequestKey | None = None,
        rate_limit_cost: CostResult | GetRequestCost | None = None,
        rate_limit_policy: RateLimitPolicy | GetRequestRateLimitPolicy | None = None,
    ) -> None:
        self.key = rate_limit_key
        self.cost = rate_limit_cost
        self.policy = rate_limit_policy
        storage_options = {**(storage_options or {})}
        if redis_uri:
            storage_options.update(
                {
                    "connection_pool": redis_connection_pool,
                    "key_prefix": redis_key_prefix,
                    "wrap_exceptions": redis_wrap_exceptions,
                }
            )
            if redis_options:
                storage_options.update(redis_options)

            if not isinstance(redis_uri, str):
                redis_uri = redis_uri.encoded_string()

            storage_uri = redis_uri

        if storage_uri:
            if not isinstance(storage_uri, str):
                storage_uri = storage_uri.encoded_string()

            storage = limits.storage.storage_from_string(storage_uri, **storage_options)
        else:
            storage = limits.storage.MemoryStorage()

        self.limiter: limits.strategies.RateLimiter
        if strategy == "fixed-window":
            self.limiter = limits.strategies.FixedWindowRateLimiter(storage)
        elif strategy == "moving-window":
            self.limiter = limits.strategies.MovingWindowRateLimiter(storage)
        elif strategy == "sliding-window-counter":
            self.limiter = limits.strategies.SlidingWindowCounterRateLimiter(storage)
        else:
            raise ValueError(
                f"Unsupported strategy: {strategy}. "
                "Available options are: 'fixed-window', 'moving-window', "
                "'sliding-window-counter'."
            )
        self.namespace = namespace


class AsyncFarl(AsyncFarlProtocol):
    @overload
    def __init__(
        self,
        *,
        redis_uri: str | RedisDsn | None,
        redis_connection_pool: "AsyncRedisConnectionPool | None" = None,  # pyright: ignore [reportInvalidTypeForm]
        redis_key_prefix: str = limits.aio.storage.RedisStorage.PREFIX,
        redis_wrap_exceptions: bool = False,
        redis_options: dict | None = None,
        strategy: Literal[
            "fixed-window",
            "moving-window",
            "sliding-window-counter",
        ] = "fixed-window",
        namespace: str | None = None,
        rate_limit_key: KeyResult | GetRequestKey | None = None,
        rate_limit_cost: CostResult | GetRequestCost | None = None,
        rate_limit_policy: RateLimitPolicy | GetRequestRateLimitPolicy | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        *,
        storage_uri: str | RedisDsn,
        storage_options: dict | None = None,
        strategy: Literal[
            "fixed-window",
            "moving-window",
            "sliding-window-counter",
        ] = "fixed-window",
        namespace: str | None = None,
        rate_limit_key: KeyResult | GetRequestKey | None = None,
        rate_limit_cost: CostResult | GetRequestCost | None = None,
        rate_limit_policy: RateLimitPolicy | GetRequestRateLimitPolicy | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        *,
        strategy: Literal[
            "fixed-window",
            "moving-window",
            "sliding-window-counter",
        ] = "fixed-window",
        namespace: str | None = None,
        rate_limit_key: KeyResult | GetRequestKey | None = None,
        rate_limit_cost: CostResult | GetRequestCost | None = None,
        rate_limit_policy: RateLimitPolicy | GetRequestRateLimitPolicy | None = None,
    ) -> None: ...

    def __init__(
        self,
        *,
        redis_uri: str | RedisDsn | None = None,
        redis_connection_pool: "AsyncRedisConnectionPool | None" = None,  # pyright: ignore [reportInvalidTypeForm]
        redis_key_prefix: str = limits.aio.storage.RedisStorage.PREFIX,
        redis_wrap_exceptions: bool = False,
        redis_options: dict | None = None,
        storage_uri: str | RedisDsn | None = None,
        storage_options: dict | None = None,
        strategy: Literal[
            "fixed-window",
            "moving-window",
            "sliding-window-counter",
        ] = "fixed-window",
        namespace: str | None = None,
        rate_limit_key: KeyResult | GetRequestKey | None = None,
        rate_limit_cost: CostResult | GetRequestCost | None = None,
        rate_limit_policy: RateLimitPolicy | GetRequestRateLimitPolicy | None = None,
    ) -> None:
        self.key = rate_limit_key
        self.cost = rate_limit_cost
        self.policy = rate_limit_policy
        storage_options = {**(storage_options or {})}
        if redis_uri:
            storage_options.update(
                {
                    "connection_pool": redis_connection_pool,
                    "key_prefix": redis_key_prefix,
                    "wrap_exceptions": redis_wrap_exceptions,
                }
            )
            if redis_options:
                storage_options.update(redis_options)

            if not isinstance(redis_uri, str):
                storage_uri = redis_uri.encoded_string()
            else:
                storage_uri = redis_uri

        if storage_uri:
            if not isinstance(storage_uri, str):
                storage_uri = storage_uri.encoded_string()
            async_storage_uri = self._to_async_storage_uri(storage_uri)

            storage = limits.storage.storage_from_string(
                async_storage_uri,
                **storage_options,
            )
        else:
            storage = limits.aio.storage.MemoryStorage()

        limiter_class: type[limits.aio.strategies.RateLimiter]
        if strategy == "fixed-window":
            limiter_class = limits.aio.strategies.FixedWindowRateLimiter
        elif strategy == "moving-window":
            limiter_class = limits.aio.strategies.MovingWindowRateLimiter
        elif strategy == "sliding-window-counter":
            limiter_class = limits.aio.strategies.SlidingWindowCounterRateLimiter

        else:
            raise ValueError(
                f"Unsupported strategy: {strategy}. "
                "Available options are: 'fixed-window', 'moving-window', "
                "'sliding-window-counter'."
            )

        self.limiter = limiter_class(cast(limits.aio.storage.Storage, storage))
        self.namespace = namespace

    @staticmethod
    def _to_async_storage_uri(uri: str):
        if uri.startswith("async+"):
            return uri
        return f"async+{uri}"



from farl.constants import HEADER_RATELIMIT, HEADER_RATELIMIT_POLICY
from farl.exceptions import FarlError


def openapi(model: type[FarlError]):
    return {
        "model": model,
        "headers": {
            HEADER_RATELIMIT_POLICY: {
                "schema": {
                    "type": "string",
                    "title": "RateLimit-Policy",
                }
            },
            HEADER_RATELIMIT: {
                "schema": {
                    "type": "string",
                    "title": "RateLimit",
                }
            },
        },
    }



import inspect
from collections.abc import Callable
from functools import wraps
from typing import Any, TypeVar

from fastapi import Depends

from farl.exceptions import FarlError, QuotaExceeded
from farl.types import (
    AnyFarlProtocol,
    CostResult,
    GetCostDependency,
    GetKeyDependency,
    GetRateLimitPolicyDependency,
    KeyResult,
    RateLimitPolicy,
)

from .dependencies import rate_limit as rate_limit_dep
from .dependencies.utils import request_endpoint


Fn = TypeVar("Fn", bound=Callable)

INJECT_DEP_PREFIX = "__navio_ratelimit"


def _exclude_kwds(kwds: dict[str, Any]):
    exclude = [i for i in kwds if i.startswith(INJECT_DEP_PREFIX)]
    for i in exclude:
        kwds.pop(i)


def _create_call(fn: Fn) -> Fn:
    if inspect.iscoroutinefunction(fn):

        @wraps(fn)
        async def async_hdlr(*args, **kwds):
            _exclude_kwds(kwds)
            return await fn(*args, **kwds)

        return async_hdlr  # pyright: ignore[reportReturnType]

    @wraps(fn)
    def hdlr(*args, **kwds):
        _exclude_kwds(kwds)
        return fn(*args, **kwds)

    return hdlr  # pyright: ignore[reportReturnType]


def rate_limit(
    argument: RateLimitPolicy | GetRateLimitPolicyDependency,
    *,
    policy_name: str | None = None,
    quota_unit: str | None = None,
    get_key: KeyResult | GetKeyDependency | None = None,
    get_partition_key: KeyResult | GetKeyDependency = request_endpoint,
    get_cost: CostResult | GetCostDependency = 1,
    error_class: type[FarlError] | None = QuotaExceeded,
    farl: AnyFarlProtocol | None = None,
):
    dep = rate_limit_dep(
        argument,
        policy_name=policy_name,
        quota_unit=quota_unit,
        get_key=get_key,
        get_partition_key=get_partition_key,
        get_cost=get_cost,
        error_class=error_class,
        farl=farl,
    )

    def decorate(fn: Fn) -> Fn:
        new_fn = _create_call(fn)

        sign = inspect.signature(new_fn)
        param_mappings = sign.parameters.copy()

        dep_no = len([i for i in param_mappings if i.startswith(INJECT_DEP_PREFIX)]) + 1
        dep_param = inspect.Parameter(
            f"{INJECT_DEP_PREFIX}_{dep_no}",
            kind=inspect.Parameter.KEYWORD_ONLY,
            default=Depends(dep),
        )
        params = list(param_mappings.values())
        if params and params[-1].kind == inspect.Parameter.VAR_KEYWORD:
            params.insert(-1, dep_param)
        else:
            params.append(dep_param)
        new_fn.__signature__ = sign.replace(parameters=params)
        return new_fn

    return decorate



import base64
from collections.abc import Sequence
from datetime import datetime
from math import ceil
from typing import TYPE_CHECKING, Literal, NamedTuple, NotRequired, TypedDict

from limits.limits import (
    RateLimitItem,
    RateLimitItemPerDay,
    RateLimitItemPerHour,
    RateLimitItemPerMinute,
    RateLimitItemPerMonth,
    RateLimitItemPerSecond,
    RateLimitItemPerYear,
)
from limits.util import parse_many

from farl.exceptions import FarlError


if TYPE_CHECKING:
    from farl.types import AnyFarlProtocol, RateLimitPolicy


_TimeUnit = Literal[
    "Y",
    "year",
    "M",
    "month",
    "D",
    "day",
    "h",
    "hour",
    "m",
    "minute",
    "s",
    "second",
]


class RateLimitDictValue(TypedDict):
    time: NotRequired[_TimeUnit]
    amount: int
    multiples: NotRequired[int]


class RateLimitTimeArg:
    __slots__ = ("time", "val")

    def __init__(
        self,
        *,
        time: _TimeUnit = "m",
        amount: int,
        multiples: int = 1,
    ) -> None:
        if time == "Y":
            time = "year"
        elif time == "M":
            time = "month"
        elif time == "D":
            time = "day"
        elif time == "h":
            time = "hour"
        elif time == "m":
            time = "minute"
        elif time == "s":
            time = "second"

        if time == "year":
            self.val = RateLimitItemPerYear(amount, multiples)
        elif time == "month":
            self.val = RateLimitItemPerMonth(amount, multiples)
        elif time == "day":
            self.val = RateLimitItemPerDay(amount, multiples)
        elif time == "hour":
            self.val = RateLimitItemPerHour(amount, multiples)
        elif time == "minute":
            self.val = RateLimitItemPerMinute(amount, multiples)
        elif time == "second":
            self.val = RateLimitItemPerSecond(amount, multiples)

        self.time = time


def parse_rate_limit_value(
    *args: "RateLimitPolicy",
):
    result: list[RateLimitItem] = []
    for i in args:
        if isinstance(i, str):
            result.extend(parse_many(i))
        elif isinstance(i, RateLimitItem):
            result.append(i)
        elif isinstance(i, RateLimitTimeArg):
            result.append(i.val)
        elif isinstance(i, dict):
            result.append(
                RateLimitTimeArg(
                    time=i.get("time", "minute"),
                    amount=i["amount"],
                    multiples=i.get("multiples", 1),
                ).val
            )
        elif isinstance(i, Sequence):
            result.extend(parse_rate_limit_value(*i))
        else:
            raise TypeError(
                f"Unsupported rate limit argument type: {type(i).__name__}."
            )

    return result


class HeaderRateLimit(NamedTuple):
    policy: str
    remaining: int
    reset_timestamp: float | None = None
    partition_key: str | bytes | None = None
    error_class: type[FarlError] | None = None

    @property
    def quota_reset_seconds(self) -> int | None:
        if self.reset_timestamp is not None:
            return ceil(self.reset_timestamp - datetime.now().timestamp())
        return None

    def __str__(self) -> str:
        values = [f'"{self.policy}"', f"r={self.remaining}"]

        if (t := self.quota_reset_seconds) is not None:
            values.append(f"t={t}")

        if self.partition_key is not None:
            if isinstance(self.partition_key, bytes):
                pk = f":{base64.b64encode(self.partition_key).decode()}:"
            else:
                pk = f'"{self.partition_key}"'

            values.append(f"pk={pk}")

        return ";".join(values)


class HeaderRateLimitPolicy(NamedTuple):
    policy: str

    quota: int
    quota_unit: str | None = None
    window: int | None = None
    partition_key: str | bytes | None = None

    def __str__(self) -> str:
        values = [f'"{self.policy}"', f"q={self.quota}"]
        if self.quota_unit is not None:
            values.append(f'qu="{self.quota_unit}"')

        if self.window is not None:
            values.append(f"w={self.window}")

        if self.partition_key is not None:
            if isinstance(self.partition_key, bytes):
                pk = f":{base64.b64encode(self.partition_key).decode()}:"
            else:
                pk = f'"{self.partition_key}"'

            values.append(f"pk={pk}")

        return ";".join(values)


class FarlState(TypedDict):
    farl: NotRequired["AnyFarlProtocol"]
    policy: list[HeaderRateLimitPolicy]
    state: list[HeaderRateLimit]
    violated: list[HeaderRateLimit]
